from typing import List, Optional, Dict
import numpy as np
from .base import Memory, MemoryStream
from .storage import LocalStorage

class MemoryManager:
    def __init__(self, importance_calculator, embedding_generator, agent):
        self.importance_calculator = importance_calculator
        self.embedding_generator = embedding_generator
        self.storage = LocalStorage(persist_directory=f"{agent.data_dir}/memories")
        self.memory_stream = MemoryStream(storage=self.storage)
        self.agent = agent

    async def add_memory(self, content: str, memory_type: str,
                        associated_agents: Optional[List[str]] = None,
                        metadata: Optional[Dict] = None) -> Memory:
        embedding = np.array(await self.embedding_generator.get_embedding(content, self.agent.get_model('embedding')))
        memory = Memory(content=content, type=memory_type,
                       associated_agents=associated_agents,
                       metadata=metadata, embedding=embedding)
        
        importance = await self.importance_calculator.calculate_memory_importance(
            memory)
        memory.importance_score = importance
        await self.memory_stream.add_memory(memory)
        return memory
    # To be used
    async def get_memory_by_id(self, memory_id: str) -> Optional[Memory]:
        return await self.memory_stream.get_memory(memory_id)
    # To be used
    async def get_memories_by_type(self, memory_type: str) -> List[Memory]:
        return await self.memory_stream.get_memories_by_type(memory_type)

    async def get_memories_by_agent(self, agent_id: str) -> List[Memory]:
        return await self.memory_stream.get_memories_by_agent(agent_id)
    
    async def get_memories_by_associated_agent(self, agent_id: str) -> List[Memory]:
        return await self.memory_stream.get_memories_by_agent(agent_id)
        


    


from typing import List, Dict, Optional, Set
from datetime import datetime
import numpy as np
import uuid
import asyncio
from functools import lru_cache
import logging
from collections import OrderedDict

logger = logging.getLogger(__name__)


class Memory:
    __slots__ = (
        "id",
        "content",
        "type",
        "associated_agents",
        "metadata",
        "embedding",
        "importance_score",
        "created_at",
        "last_accessed",
    )

    def __init__(
    self,
    content: str,
    type: str = "observation",
    associated_agents: Optional[List[str]] = None,
    metadata: Optional[Dict] = None,
    embedding: Optional[np.ndarray] = None,
    importance_score: float = 0.0,
    id: Optional[str] = None,
    created_at: Optional[datetime] = None,
    last_accessed: Optional[datetime] = None,
):
        self.id = id or str(uuid.uuid4())
        self.content = content
        self.type = type
        self.associated_agents = associated_agents or []
        self.metadata = metadata or {}
        self.embedding = embedding
        self.importance_score = importance_score
        self.created_at = created_at or datetime.now()
        self.last_accessed = last_accessed or datetime.now()

    @lru_cache(maxsize=1000)
    async def to_dict(self) -> Dict:
        return {
            "id": self.id,
            "content": self.content,
            "type": self.type,
            "associated_agents": self.associated_agents,
            "metadata": self.metadata,
            "embedding": (
                self.embedding.tolist() if self.embedding is not None else None
            ),
            "importance_score": self.importance_score,
            "created_at": self.created_at,
            "last_accessed": self.last_accessed,
        }

    @classmethod
    async def from_dict(cls, data: Dict):
        if data.get("embedding") is not None:
            data["embedding"] = np.array(data["embedding"])
        return cls(**data)


class MemoryStream:
    def __init__(self, storage=None, cache_size: int = 1000):
        self.memories: OrderedDict[str, Memory] = OrderedDict()
        self._cache: OrderedDict[str, Memory] = OrderedDict()
        self._save_queue: asyncio.Queue[Memory] = asyncio.Queue()
        self.storage = storage
        self._initialized = False
        self._initialization_lock = asyncio.Lock()
        self._cache_size = cache_size
        self._batch_size = 100
        self._save_task = None

    async def _start_save_worker(self):
        while True:
            batch = []
            try:
                while len(batch) < self._batch_size:
                    memory = await self._save_queue.get()
                    batch.append(memory)
                    if self._save_queue.empty():
                        break
                if batch:
                    await self._batch_save(batch)
            except Exception as e:
                logger.error(f"Error in save worker: {e}")
                await asyncio.sleep(1)

    async def _batch_save(self, memories: List[Memory]):
        if self.storage:
            tasks = [self.storage.save_memory(memory) for memory in memories]
            await asyncio.gather(*tasks)

    def _update_cache(self, memory_id: str, memory: Memory):
        self._cache[memory_id] = memory
        if len(self._cache) > self._cache_size:
            self._cache.popitem(last=False)

    async def ensure_initialized(self):
        async with self._initialization_lock:
            if not self._initialized:
                if self.storage:
                    await self.storage.initialize()
                await self.initialize()
                self._initialized = True
                if self.storage:
                    self._save_task = asyncio.create_task(self._start_save_worker())

    async def initialize(self):
        if self.storage:
            try:
                memories = await self.storage.list_all_memories()

                tasks = []
                for memory in memories:
                    tasks.append(self._process_memory(memory))
                await asyncio.gather(*tasks)
            except Exception as e:
                logger.error(f"Failed to initialize memory stream: {e}")
                self.memories = OrderedDict()

    async def _process_memory(self, memory: Memory):
        try:
            memory_dict = await memory.to_dict()
            if isinstance(memory_dict, dict):
                self.memories[memory.id] = memory
                self._update_cache(memory.id, memory)
        except Exception as e:
            logger.error(f"Failed to process memory: {e}")

    async def add_memory(self, memory: Memory) -> str:
        await self.ensure_initialized()
        self.memories[memory.id] = memory
        self._update_cache(memory.id, memory)
        if self.storage:
            await self._save_queue.put(memory)
        return memory.id

    async def get_memory(self, memory_id: str) -> Optional[Memory]:
        if memory_id in self._cache:
            memory: Optional[Memory] = self._cache[memory_id]
        else:
            memory = self.memories.get(memory_id)
            if memory is None and self.storage:
                memory = await self.storage.load_memory(memory_id)
                if memory:
                    self.memories[memory_id] = memory
                    self._update_cache(memory_id, memory)
        if memory:
            memory.last_accessed = datetime.now()
            if self.storage:
                await self._save_queue.put(memory)
        return memory

    async def update_memory(self, memory: Memory) -> None:
        if memory.id in self.memories:
            self.memories[memory.id] = memory
            self._update_cache(memory.id, memory)
            if self.storage:
                await self.storage.save_memory(memory)

    async def delete_memory(self, memory_id: str) -> bool:
        if memory_id in self.memories:
            del self.memories[memory_id]
            self._cache.pop(memory_id, None)
            if self.storage:
                await self.storage.delete_memory(memory_id)
            return True
        return False

    async def clear(self) -> None:
        self.memories.clear()
        self._cache.clear()
        if self.storage:
            await self.storage.clear_all()

    async def get_all_memories(self):
        await self.ensure_initialized()
        return list(self.memories.values())

    async def get_memories_by_type(self, memory_type: str) -> List[Memory]:
        await self.ensure_initialized()
        return [
            memory for memory in self.memories.values()
            if memory.type == memory_type
        ]
    
    async def get_memories_by_associated_agent(self, agent_id: str) -> List[Memory]:
        await self.ensure_initialized()
        return [
            memory for memory in self.memories.values()
            if agent_id in (memory.associated_agents or [])
        ]

    async def update_memory_immediately(self, memory: Memory) -> None:
        if memory.id in self.memories:
            self.memories[memory.id] = memory
            self._update_cache(memory.id, memory)
            if self.storage:
                await self.storage.save_memory_immediately(memory)



from typing import Dict, Any, Optional, List, Deque
import json
import os
from datetime import datetime
import numpy as np
import aiofiles
import asyncio
from collections import deque
from .base import Memory


class LocalStorage:
    def __init__(self, persist_directory: str, max_batch_size: int = 100, max_open_files: int = 256):
        self.persist_directory = persist_directory
        self.max_batch_size = max_batch_size
        self._write_queue: asyncio.Queue[Memory] = asyncio.Queue()
        self._write_lock : asyncio.Lock = asyncio.Lock()
        self._memory_cache : Dict[str, Memory] = {}
        self._write_batch : Deque[Memory] = deque()
        self._write_task: Optional[asyncio.Task] = None
        os.makedirs(persist_directory, exist_ok=True)
        self._file_semaphore = asyncio.Semaphore(max_open_files)

    async def initialize(self):
        if self._write_task is None:
            self._write_task = asyncio.create_task(self._start_write_worker())

    async def _start_write_worker(self):
        while True:
            try:
                memory = await self._write_queue.get()
                self._write_batch.append(memory)

                if (
                    len(self._write_batch) >= self.max_batch_size
                    or self._write_queue.empty()
                ):
                    async with self._write_lock:
                        batch = list(self._write_batch)
                        self._write_batch.clear()
                        await self._batch_write(batch)
            except Exception as e:
                print(f"Error in write worker: {e}")
                await asyncio.sleep(1)

    async def _batch_write(self, memories: List[Memory]):
        for memory in memories:
            memory_dict = await memory.to_dict()
            serialized_dict = self._serialize_memory(memory_dict)
            path = self._get_memory_path(memory.id)
            async with self._file_semaphore:
                async with aiofiles.open(path, "w") as f:
                    await f.write(json.dumps(serialized_dict))
            self._memory_cache[memory.id] = memory

    def _get_memory_path(self, memory_id: str) -> str:
        return os.path.join(self.persist_directory, f"{memory_id}.json")

    def _serialize_memory(self, memory_dict: Dict[str, Any]) -> Dict[str, Any]:
        serialized = memory_dict.copy()
        serialized["created_at"] = memory_dict["created_at"].isoformat()
        serialized["last_accessed"] = memory_dict["last_accessed"].isoformat()
        if memory_dict["embedding"] is not None:
            if isinstance(memory_dict["embedding"], np.ndarray):
                serialized["embedding"] = memory_dict["embedding"].tolist()
            else:
                serialized["embedding"] = memory_dict["embedding"]
        return serialized

    def _deserialize_memory(self, memory_dict: Dict[str, Any]) -> Dict[str, Any]:
        deserialized = memory_dict.copy()
        deserialized["created_at"] = datetime.fromisoformat(memory_dict["created_at"])
        deserialized["last_accessed"] = datetime.fromisoformat(
            memory_dict["last_accessed"]
        )
        if memory_dict["embedding"] is not None:
            deserialized["embedding"] = np.array(memory_dict["embedding"])
        return deserialized

    async def save_memory(self, memory: Memory) -> None:
        await self._write_queue.put(memory)

    async def load_memory(self, memory_id: str) -> Optional[Memory]:
        if memory_id in self._memory_cache:
            return self._memory_cache[memory_id]

        memory_path = self._get_memory_path(memory_id)
        if not os.path.exists(memory_path):
            return None

        async with self._file_semaphore:
            async with aiofiles.open(memory_path, "r") as f:
                content = await f.read()
                memory_dict = json.loads(content)

        deserialized_dict = self._deserialize_memory(memory_dict)
        memory = await Memory.from_dict(deserialized_dict)
        self._memory_cache[memory_id] = memory
        return memory

    async def delete_memory(self, memory_id: str) -> bool:
        memory_path = self._get_memory_path(memory_id)
        if os.path.exists(memory_path):
            os.remove(memory_path)
            self._memory_cache.pop(memory_id, None)
            return True
        return False

    async def list_all_memories(self) -> List[Memory]:
        tasks = []
        for filename in os.listdir(self.persist_directory):
            if filename.endswith(".json"):
                memory_id = filename[:-5]
                tasks.append(self.load_memory(memory_id))

        memories = await asyncio.gather(*tasks)
        return [m for m in memories if m is not None]

    async def clear_all(self) -> None:
        for filename in os.listdir(self.persist_directory):
            if filename.endswith(".json"):
                os.remove(os.path.join(self.persist_directory, filename))
        self._memory_cache.clear()

    async def save_memory_immediately(self, memory: Memory) -> None:
        memory_dict = await memory.to_dict()
        serialized_dict = self._serialize_memory(memory_dict)
        memory_path = self._get_memory_path(memory.id)
        
        async with self._file_semaphore:
            async with aiofiles.open(memory_path, "w") as f:
                await f.write(json.dumps(serialized_dict))
        
        self._memory_cache[memory.id] = memory



import os
import json
import requests
from datetime import datetime, timedelta

class DataTypes:
    MODEL_RESPONSE = 'model_response'
    SERVER_MESSAGE = 'server_message'
    SYSTEM_MESSAGE = 'system_message'

def load_file(filename):
    with open(filename, 'r') as f:
        return json.load(f)

def save_file(content, filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w') as f:
        f.write(content)

def save_image(url, filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    response = requests.get(url)
    response.raise_for_status()
    with open(filename, 'wb') as f:
        f.write(response.content)

def save_data(data_dir, name, data_type, data, memory_manager=None):
    os.makedirs(data_dir, exist_ok=True)
    with open(f'{data_dir}/{name}.jsonl', 'a') as f:
        json.dump({
            'type': data_type,
            'data': data,
            'timestamp': datetime.now().isoformat()
        }, f)
        f.write('\n')
    
async def save_to_memory(data, memory_manager):
    if not memory_manager:
        return
    
    msg_type = data.get('type')
    if msg_type == 'actionResult':
        action_data = data.get('data', {})
        action_result = f"{action_data.get('payload', {})}"
        action_id = action_data.get('actionID')
        
        all_memories = await memory_manager.memory_stream.get_all_memories()
        for memory in all_memories:
            metadata = memory.metadata
            if not memory.metadata:
                memory.metadata = {}
                metadata = memory.metadata
            if memory.type == 'actionResult' and metadata.get('action_id') == action_id:
                try:  
                    content_json = json.loads(metadata.get('original_content').rstrip('.'))
                    content_json['actionResult'] = action_result
                    metadata['original_content']= json.dumps(content_json)
                    await memory_manager.memory_stream.update_memory(memory)
                except json.JSONDecodeError:
                    print(f"Error: Could not parse original_content as JSON: {memory.original_content}")
                break
    elif msg_type == 'chat':
        chat_data = data.get('data', {})
        content = f"{chat_data.get('message')}"
        metadata = {
            'sender_id': str(chat_data.get('senderPlayerID')),
        }
        await memory_manager.add_memory(
            content=content,
            memory_type='chat',
            metadata=metadata,
            associated_agents=[chat_data.get('senderPlayerID')]
        )
    elif msg_type == 'system':
        system_data = data.get('data', {})
        content = f"{system_data.get('message')}"
        channel_id = system_data.get('channelID')
        metadata = {
            'channel_id': channel_id,
        }
        await memory_manager.add_memory(
            content=content,
            memory_type='system',
            metadata=metadata,
            associated_agents=[]
        )

def load_data(data_dir, name, data_type=None, since=None, max_responses=None):
    with open(f'{data_dir}/{name}.jsonl', 'r') as f:
        data = [json.loads(line) for line in f if line.strip()]
    
    if data_type is not None:
        data = [d for d in data if d['type'] == data_type]
    
    if since is not None:
        if isinstance(since, (int, float)):
            since_dt = datetime.now() - timedelta(hours=since)
        elif isinstance(since, str):
            since_dt = datetime.fromisoformat(since)
        else:
            since_dt = since
        data = [d for d in data if datetime.fromisoformat(d['timestamp']) >= since_dt]
    
    if max_responses is not None:
        data = data[-max_responses:]

    return data



from typing import List, Dict, Optional, Any, TYPE_CHECKING
from dataclasses import dataclass
from datetime import datetime
from .base import Memory
from .reflection import MemoryReflection
from .manager import MemoryManager
import json

if TYPE_CHECKING:
    from ..agent import Agent

@dataclass
class MemoryStep:
    thought: Optional[str] = None
    action: Optional[str] = None
    action_input: Optional[str] = None
    action_result: Optional[str] = None  
    observation: Optional[str] = None
    timestamp: Optional[datetime] = None
    metadata: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        if self.timestamp is None:
            self.timestamp = datetime.now()
        if self.metadata is None:
            self.metadata = {}

    def to_string(self) -> str:
        content_dict = {
            "thought": self.thought if self.thought else "",
            "action": self.action if self.action else "",
            "actionInput": self.action_input if self.action_input else "",
            "actionResult": self.action_result if self.action_result else "",
            "observation": self.observation if self.observation else ""
        }
        return json.dumps(content_dict)

    async def to_memory(self) -> Memory:
        metadata_copy = {}
        if self.metadata is not None:
            metadata_copy = self.metadata.copy()
            metadata_copy.pop('type', None)
            memory_type = self.metadata.get('type', '')
        else:
            memory_type = ''  # Default value when metadata is None
            
        return Memory(
            content=self.to_string(),
            type=memory_type,
            created_at=self.timestamp,
            metadata=metadata_copy,
        )


class WorkingMemory:
    def __init__(
        self,
        max_size: int = 10,
        agent: Optional["Agent"] = None,
        memory_manager: Optional[MemoryManager] = None,
    ):
        self.max_size = max_size
        self.steps : List[MemoryStep] = []
        self.memory_manager = memory_manager
        self.agent = agent
        self.memory_reflection = MemoryReflection(agent=agent)

    def __len__(self) -> int:
        return len(self.steps)

    async def _step_to_memory(self, step: MemoryStep) -> Memory:
        metadata = {}
        if hasattr(step, 'metadata') and isinstance(step.metadata, dict):
            metadata = step.metadata
        
        return Memory(
            content=step.to_string(),
            type="working_memory",
            metadata=metadata,
            created_at=datetime.now()
        )

    async def add_step(
        self,
        thought: Optional[str] = None,
        action: Optional[str] = None,
        action_input: Optional[str] = None,
        action_result: Optional[str] = None,  
        observation: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> MemoryStep:
        step = MemoryStep(
            thought=thought,
            action=action,
            action_input=action_input,
            action_result=action_result, 
            observation=observation,
            metadata=metadata,
        )

        self.steps.append(step)

        return step

    # To be used
    def get_recent_steps(self, n: Optional[int] = None) -> List[MemoryStep]:
        if n is None:
            return self.steps.copy()
        return self.steps[-n:]

    # To be used
    def get_last_step(self) -> Optional[MemoryStep]:
        return self.steps[-1] if self.steps else None

    def clear(self) -> None:
        self.steps.clear()

    async def _compress_steps(self) -> None:
        if len(self.steps) <= self.max_size or not self.memory_manager:
            return

        memories_to_compress = []
        for step in self.steps[-self.max_size :]:
            memory = await self._step_to_memory(step)
            memories_to_compress.append(memory)

        for memory in memories_to_compress:
            compressed_content = await self.memory_reflection.compress_memory_content(
                memory
            )
            metadata = memory.metadata or {}
            metadata.update(
                {
                    "compressed": True,
                    "original_content": memory.content,
                }
            )
            await self.memory_manager.add_memory(
                content=compressed_content,
                memory_type=memory.type,
                associated_agents=memory.associated_agents or [],
                metadata=metadata,
            )
        self.steps = []

    def steps_to_string(self) -> str:
        if not self.steps:
            return "No steps recorded."

        formatted_steps = []
        for i, step in enumerate(self.steps, 1):
            step_str = f"Step {i}:\n{step.to_string()}"
            formatted_steps.append(step_str)

        return "\n\n".join(formatted_steps)



from typing import List, Optional
from datetime import datetime
import numpy as np
from .base import Memory
import logging

logger = logging.getLogger(__name__)

class MemoryReflection:
    def __init__(self, agent):
        self.agent = agent

    # To be used
    def _filter_by_time(self, memories: List[Memory],
                       start_time: Optional[datetime] = None,
                       end_time: Optional[datetime] = None) -> List[Memory]:
        if not memories:
            return []
        if not start_time and not end_time:
            return memories
            
        filtered = memories.copy()
        if start_time:
            filtered = [m for m in filtered if m.created_at >= start_time]
        if end_time:
            filtered = [m for m in filtered if m.created_at <= end_time]
            
        return filtered

    # To be used
    def format_memories(self, memories: List[Memory]) -> List[Memory]:
        formatted_memories = []
        for memory in memories:
            formatted_memory = Memory(
                content=self._format_memory_content(memory.content),
                type=memory.type,
                created_at=memory.created_at,
                embedding=np.array([0.1] * 1536)
            )
            formatted_memories.append(formatted_memory)
        return formatted_memories
    # To be used
    def generate_summary(self, memories: List[Memory], 
                        start_time: Optional[datetime] = None,
                        end_time: Optional[datetime] = None) -> str:
        if not memories:
            return "No memories to analyze"
            
        filtered_memories = self._filter_by_time(memories, start_time, end_time)
        if not filtered_memories:
            return "No memories in the specified time period"
            
        summary_points = []
        for memory in filtered_memories:
            content = memory.content
            if hasattr(memory, 'type'):
                summary_points.append(f"- {content} (Type: {memory.type})")
            else:
                summary_points.append(f"- {content}")
            
        return "\n".join(summary_points)


    def _format_memory_content(self, content: str) -> str:
        if not content:
            return ""
        
        parts = []
        current = ""
        for char in content:
            if char in ".!?":
                current += char
                if current.strip():
                    parts.append(current.strip())
                current = ""
            else:
                current += char
        if current.strip():
            parts.append(current.strip())
        
        formatted_parts = []
        for part in parts:
            words = part.split()
            if words:
                words[0] = words[0].capitalize()
                formatted = " ".join(words)
                if not formatted[-1] in ".!?":
                    formatted += "."
                formatted_parts.append(formatted)
            
        return " ".join(formatted_parts)

    async def compress_memory_content(self, memory: Memory, max_length: int = 500) -> str:
        try:
            response = await self.agent.get_model_response(
                'agent.compress_memory',
                memory_content=memory.content,
                max_length=max_length 
            )
            
            compressed = response.get('compressed_content', '')
            return compressed[:min(len(compressed), max_length)]
        except Exception as e:
            logger.error(f"Error compressing memory content: {e}")
            return memory.content[:max_length]


from typing import List, Union
from datetime import datetime
import numpy as np
from .base import Memory
import logging

logger = logging.getLogger(__name__)

class ImportanceCalculator:
    def __init__(self, agent):
        self.time_weight = 0.3
        self.relevance_weight = 0.7
        self.agent = agent

    async def calculate_relevance(self, memory: Memory, current_time: datetime,
                                related_memories: List[Memory]) -> tuple[List[float], List[float]]:
        memory_times = [m.created_at for m in related_memories]
        time_factor = self._calculate_time_decay(memory.created_at, memory_times)
        relevance_factor = await self._calculate_relevance(memory, related_memories)
        return time_factor, relevance_factor

    async def calculate_memory_importance(self, memory: Memory) -> float:
        try:
            response = await self.agent.get_model_response(
                'agent.importance',
                memory_content=memory.content,
            )

            importance_score = response.get("importance_score")
            if importance_score is None:
                logger.error("Missing importance_score in response")
                return 0.05

            try:
                importance_score = float(importance_score)
            except (ValueError, TypeError):
                logger.error("Invalid importance_score format")
                return 0.05

            if "reasoning" in response:
                logger.debug(f"Memory importance calculation: {response['reasoning']}")
            return max(0.0, min(1.0, importance_score))
        
        except Exception as e:
            logger.error(f"Error calculating memory importance: {str(e)}")
            return 0.05

    def _calculate_time_decay(self, memory_created_at: datetime, 
                         reference_times: Union[datetime, List[datetime]]) -> List[float]:
        if isinstance(reference_times, datetime):
            time_diff = (reference_times - memory_created_at).total_seconds()
            return [abs(1.0 / (1.0 + time_diff / (24 * 3600)))]
        
        time_decays = []
        for ref_time in reference_times:
            time_diff = (ref_time - memory_created_at).total_seconds()
            decay = abs(1.0 / (1.0 + time_diff / (24 * 3600)))
            time_decays.append(decay)
            
        return time_decays

    async def _calculate_relevance(self, memory: Memory, related_memories: List[Memory]) -> List[float]:
        if not related_memories or memory.embedding is None:
            return [0.1] 
        similarities = []
        for related_memory in related_memories:
            if related_memory.embedding is not None:
                similarity = np.dot(memory.embedding, related_memory.embedding)
                similarities.append(similarity)
        
        return similarities if similarities else [0.05] * len(related_memories)



from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Optional
if TYPE_CHECKING:
    from .agent import Agent

import asyncio
import json
import logging
import time
import os
import websockets
from datetime import datetime
from typing import List
from .data import DataTypes
from .utils import (
    distance_obj,
    is_valid_state_data,
    is_valid_map_data,
    is_valid_player_data,
    remove_additional_data,
    seconds_since,
    format_json,
    format_memory,
)
from .memory.base import Memory

logger = logging.getLogger(__name__)

class MessagingHandler:
    agent: "Agent"

    def __init__(self, agent):
        self.agent = agent
        self.message_queue = asyncio.Queue()
        self.map_data = None
        self.nearby_map = None
        self.players_data = None
        self.nearby_players = None
        self.assets_data = None
        self.inventory_data = None
        self.state_data = None
        self.available_actions = None
        self.system_messages = []
        self.other_data = []
        self.last_prompt_time = None
        self.use_model = False
        self.min_model_interval = int(os.getenv('MIN_MODEL_INTERVAL', 5))
        self.max_model_interval = int(os.getenv('MAX_MODEL_INTERVAL', 60))
        self.model_interval = self.min_model_interval
        self.action_id = int(time.time() * 1000)

    async def connect(self, uri):
        self.websocket = await websockets.connect(uri)
        return self

    async def __aenter__(self):
        return self.websocket

    async def __aexit__(self, exc_type, exc, tb):
        await self.websocket.close()

    async def handle_messages(self, websocket, stop_event):
        receive_task = asyncio.create_task(self._receive_messages(websocket, stop_event))
        process_task = asyncio.create_task(self._process_messages(websocket, stop_event))

        done, pending = await asyncio.wait(
            [receive_task, process_task],
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in done:
            if task.exception():
                raise task.exception()

        for task in pending:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.info(f"Task {task} has been cancelled successfully.")

    async def _receive_messages(self, websocket, stop_event):
        try:
            async for message in websocket:
                if stop_event.is_set():
                    logger.info("Stop event detected. Exiting handle_messages.")
                    break
                await self.message_queue.put(json.loads(message))
        except websockets.exceptions.ConnectionClosed as e:
            logger.info(f"Receive messages - Connection closed: {e}")
            raise

    async def _process_messages(self, websocket, stop_event):
        try:
            while True:
                if stop_event.is_set():
                    logger.info("Stop event detected. Exiting handle_messages.")
                    break

                try:
                    message = await asyncio.wait_for(self.message_queue.get(), timeout=60)
                    if stop_event.is_set():
                        logger.info("Stop event detected. Exiting handle_messages.")
                        break
                    if not await self._process_message(message):
                        continue
                except asyncio.TimeoutError:
                    self.use_model = True
                    message = None
                except asyncio.CancelledError:
                    raise

                if not self.message_queue.empty():
                    continue

                if len(self.system_messages) == 0:
                    seconds_since_last_prompt = seconds_since(self.last_prompt_time) if self.last_prompt_time else self.max_model_interval
                    if seconds_since_last_prompt < self.model_interval:
                        continue
                    if seconds_since_last_prompt < self.max_model_interval and not self.use_model:
                        continue

                try:
                    response = await self._generate_model_response()
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    logger.error(f"Error generating model response: {e}")
                    continue

                try:
                    await self._process_model_response(response)
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    logger.error(f"Error processing model response: {e}")
                    continue

                try:
                    await self._send_action(response, websocket)
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    logger.error(f"Error sending action: {e}")
                    raise
        except asyncio.CancelledError:
            logger.info("Process messages has been cancelled.")
            raise
        except Exception as e:
            logger.error(f"Process messages encountered an unexpected error: {e}")
            raise

    async def _process_message(self, message):
        msg_type = message.get('type')
        data = message.get('data')

        if msg_type not in ['tickEnd', 'map', 'players']:
            logger.info(f"Received: {msg_type} {data}")

        if msg_type == 'map':
            self.map_data = data
        elif msg_type == 'players':
            self.players_data = data
        elif msg_type == 'assets':
            self.assets_data = data
            self.use_model = True
        elif msg_type == 'inventory':
            self.inventory_data = data
            self.use_model = True
        elif msg_type == 'state':
            self.state_data = data
            if self.state_data['state'] in ['moving', 'learning', 'working']:
                self.model_interval = self.max_model_interval if self.state_data['state'] != 'moving' else self.model_interval
            else:
                self.model_interval = self.min_model_interval
            self.use_model = True
        elif msg_type == 'availableActions':
            self.available_actions = data
            self.use_model = True
        elif msg_type == 'tickEnd':
            pass
        elif msg_type == 'system':
            self.system_messages.append(message)
            self.use_model = True
            await self.agent._save_data(DataTypes.SYSTEM_MESSAGE, message)
        else:
            self.other_data.append(message)
            self.use_model = True
            await self.agent._save_data(DataTypes.SERVER_MESSAGE, message)

        if is_valid_state_data(self.state_data) and is_valid_map_data(self.map_data, 'buildings'):
            nearby_map = {
                'buildings': [],
                'smart_buildings': [],
                'my_houses': []
            }

            for building in self.map_data['buildings']:
                building['distance'] = distance_obj(building['entrance'], self.state_data)

            sorted_buildings = sorted(
                self.map_data['buildings'],
                key=lambda b: b.get('distance', float('inf'))
            )

            for building in sorted_buildings:
                remove_additional_data(building)
                distance = building.get('distance', float('inf'))
                if distance > self.agent.vision_range_buildings and len(nearby_map['buildings']) >= self.agent.min_num_buildings:
                    break
                nearby_map['buildings'].append(building)

            for building in sorted_buildings:
                remove_additional_data(building)
                if len(building.get('smartActions', {})) > 0:
                    if isinstance(building.get('smartActions'), dict):
                        building['smartActions'] = list(building.get('smartActions', {}).keys())
                    nearby_map['smart_buildings'].append(building)

            if is_valid_map_data(self.assets_data, 'rentedBuildings'):
                for building in self.assets_data['rentedBuildings']:
                    building['distance'] = distance_obj(building['entrance'], self.state_data)

                sorted_rented_buildings = sorted(
                    self.assets_data['rentedBuildings'],
                    key=lambda b: b.get('distance', float('inf'))
                )

                for building in sorted_rented_buildings:
                    remove_additional_data(building)
                    distance = building.get('distance', float('inf'))
                    if distance > self.agent.vision_range_buildings and len(nearby_map['my_houses']) >= self.agent.min_num_buildings:
                        break
                    nearby_map['my_houses'].append(building)

            self.nearby_map = nearby_map

        if is_valid_state_data(self.state_data) and is_valid_player_data(self.players_data):
            nearby_players = []
            for player in self.players_data:
                player['distance'] = distance_obj(player, self.state_data)

            sorted_players = sorted(
                self.players_data,
                key=lambda p: p.get('distance', float('inf'))
            )

            for player in sorted_players:
                if player.get('playerID') == self.state_data.get('playerID'):
                    continue
                distance = player.get('distance', float('inf'))
                if distance > self.agent.vision_range_players and len(nearby_players) >= self.agent.min_num_players:
                    break
                nearby_players.append(player)

            self.nearby_players = nearby_players

        return msg_type == 'tickEnd'

    async def get_working_and_long_term_memories(self, recent_steps_limit: int = 1) -> tuple[str, str]:
        recent_steps = self.agent.working_memory.get_recent_steps(recent_steps_limit)
        
        if recent_steps:
            working_memory_content = "\n".join(step.to_string() for step in recent_steps)
        else:
            working_memory_content = ""
        all_memories = await self.agent.memory_manager.memory_stream.get_all_memories()
        long_term_memory = await self._filter_and_rank_memories(working_memory_content, all_memories)
        
        return working_memory_content, long_term_memory

    async def _get_chat_context(self, sender_id: str, limit: int = 5) -> List[Memory]:
        chat_memories = await self.agent.memory_manager.memory_stream.get_memories_by_type('chat')
        relevant_memories = [
            memory for memory in chat_memories
            if (str(sender_id) in memory.metadata['sender_id'])
        ]
        
        return sorted(relevant_memories, key=lambda x: x.created_at, reverse=True)[:limit]

    async def _get_system_context(self, limit: int = 5) -> List[Memory]:
        system_memories = await self.agent.memory_manager.memory_stream.get_memories_by_type('system')
        return sorted(system_memories, key=lambda x: x.created_at, reverse=True)[:limit]

    async def _merge_memories_to_string(self, existing_memory: str, new_memories: List[Memory]) -> str:
        if not isinstance(existing_memory, str):
            existing_memory = ""
            
        memory_contents = [memory.content for memory in new_memories]
        return existing_memory + "\n" + "\n".join(memory_contents) if existing_memory else "\n".join(memory_contents)

    async def _process_chat_message(self, long_term_memory: str, current_message: Dict) -> str:
        sender_id = current_message.get('data', {}).get('senderID')
        if sender_id:
            chat_context = await self._get_chat_context(sender_id)
            merged_memory = await self._merge_memories_to_string(long_term_memory, chat_context)
            return merged_memory

        return long_term_memory

    async def _process_system_message(self, long_term_memory: str, current_message: Dict) -> str:
        channel_id = current_message.get('data', {}).get('channelID')
        if channel_id:
            system_context = await self._get_system_context(channel_id)
            merged_memory = await self._merge_memories_to_string(long_term_memory, system_context)
            return merged_memory
        return long_term_memory

    async def _generate_model_response(self):
        working_memory_content, long_term_memory = await self.get_working_and_long_term_memories(10)
        for msg in self.other_data:
            if msg.get('type') == 'chat':
                long_term_memory = await self._process_chat_message(long_term_memory, msg)
            
        for msg in self.system_messages:
            if msg.get('type') == 'system':
                long_term_memory = await self._process_system_message(long_term_memory, msg)
        
        character_info = self.agent.get_prompt('character.info')
        system_prompt = self.agent.get_prompt('agent.system').format(
            character_name=self.agent.name,
            character_info=character_info,
        )
        return await self.agent.get_model_response(
            'agent.agent',
            character_name=self.agent.name,
            character_info=character_info,
            nearby_map=format_json(self.nearby_map),
            nearby_players=format_json(self.nearby_players),
            assets=format_json(self.assets_data),
            inventory=format_json(self.inventory_data),
            state=format_json(self.state_data),
            available_actions=format_json(self.available_actions),
            working_memory=self.agent.working_memory.steps_to_string(),
            long_term_memory=long_term_memory,
            planning=self.agent.planning,
            system_messages=format_json(self.system_messages),
            messages=format_json(self.other_data),
            current_time=datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
            system_prompt=system_prompt,
        )

    async def _process_model_response(self, response):
        await self.agent._save_data(DataTypes.MODEL_RESPONSE, response)

        logger.info('-' * 100)
        logger.info(f"Observation: {response['observation']}")
        logger.info(f"Thought: {response['thought']}")
        logger.info(f"Planning: {response.get('planning', self.agent.planning)}")
        logger.info(f"Action: {response['action']}")
        logger.info(f"System message reply action: {response.get('systemMessageReplyAction')}")
        logger.info('-' * 100)

        new_planning = response.get('planning', '')
        if new_planning:
            self.agent.planning = new_planning

        await self._add_working_memory(response, self.state_data)

        logger.info(f"Working memory size: {len(self.agent.working_memory)}")

        await self.agent.working_memory._compress_steps()

        self.system_messages = []
        self.other_data = []
        self.last_prompt_time = datetime.now()
        self.use_model = False
        self.model_interval = self.min_model_interval

    async def _add_working_memory(self, response: Dict, state_data: Dict) -> None:
        thought = response.get('thought')
        action = response.get('action')
        action_input = response.get('actionInput')
        observation = response.get('observation')
        action_id = getattr(action, 'get', lambda x: None)('actionID')

        metadata = {
            'type': 'response',
            'action_id': action.get('actionID') if action else None,
        }
        metadata.update({
            "location": {
                "locationX": state_data.get('locationX'),
                "locationY": state_data.get('locationY')
            }
        })
        await self.agent.working_memory.add_step(
            thought=thought,
            action=action,
            action_input=action_input,
            action_result="",
            observation=observation,
            metadata=metadata
        )

    async def _send_action(self, response, websocket):
        if not response.get('action') or response['action'].get('action') == 'no action':
            logger.info('No action to perform.')
        else:
            response['action']['actionID'] = self.action_id
            self.action_id += 1
            logger.info(f"Sending action: {response['action']}")
            await websocket.send(json.dumps(response['action']))

        if response.get('systemMessageReplyAction'):
            logger.info(f"Sending system message reply action: {response['systemMessageReplyAction']}")
            await websocket.send(json.dumps(response['systemMessageReplyAction']))

    async def _filter_and_rank_memories(self, working_memory_content: str, all_memories: List[Memory]) -> str:
        if not (working_memory_content and all_memories and self.state_data):
            return ""

        current_memory = Memory(
            content=working_memory_content,
            type="working_memory",
            created_at=datetime.now()
        )
        current_memory_embedding = await self.agent.embedding_generator.get_embedding(current_memory.content, self.agent.get_model('embedding'))
        current_memory.embedding = current_memory_embedding
        
        time_factor, relevance_factor = await self.agent.importance_calculator.calculate_relevance(
            current_memory,
            datetime.now(),
            all_memories
        )

        memory_importance = [
            min(max(self.agent.importance_calculator.time_weight * t + self.agent.importance_calculator.relevance_weight * r, 0.0), 1.0)
            for t, r in zip(time_factor, relevance_factor)
        ]

        current_pos = (self.state_data.get('locationX', 0), self.state_data.get('locationY', 0))
        filtered_indices = []
        for i, memory in enumerate(all_memories):
            if not memory.metadata or 'location' not in memory.metadata:
                continue
            memory_pos = (
                memory.metadata['location'].get('locationX', 0),
                memory.metadata['location'].get('locationY', 0)
            )
            
            distance = abs(current_pos[0] - memory_pos[0]) + abs(current_pos[1] - memory_pos[1])
            if distance <= self.agent.vision_range_buildings:
                filtered_indices.append(i)
        if filtered_indices:
            filtered_importance = [memory_importance[i] for i in filtered_indices]
            top_k = min(3, len(filtered_indices))
            top_indices = sorted(range(len(filtered_importance)), 
                            key=lambda i: filtered_importance[i],
                            reverse=True)[:top_k]
            top_memories = [all_memories[filtered_indices[i]] for i in top_indices]
        else:
            top_k = min(3, len(all_memories))
            top_indices = sorted(range(len(memory_importance)),
                            key=lambda i: memory_importance[i],
                            reverse=True)[:top_k]
            top_memories = [all_memories[i] for i in top_indices]

        return "\n".join(memory.content for memory in top_memories)


from typing import List, Optional
import logging
from tenacity import retry, stop_after_attempt, wait_exponential

logger = logging.getLogger(__name__)

class EmbeddingGenerator:
    def __init__(self, model_manager):
        self.dimension = 1536  
        self.model_manager = model_manager

    @retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10))
    async def get_embedding(self, text: str, model: Optional[str] = None) -> List[float]:
        try:
            response = await self.model_manager.embedding(
                model=model,
                input=text,
            )
            return response['data'][0]['embedding']
        except Exception as e:
            logger.error(f"Error generating embedding: {e}")
            return [0.0] * self.dimension
    